package plaGen;

import java.util.Date;
import java.util.Vector;

/*
 * Class Name: Cell
 * 
 * Constructors:
 * Cell(Caseinfo,Paramgetter):  Creates a Cell object based off the information
 * contained in the Caseinfo and Paramgetter Objects.
 * 1. Sets up parameter (params) and case-statement (info) data for use in
 * cell generation.
 * 2. Initializes node, arc and export storage vectors, which will eventually
 * contain all cell data.
 * 3. Calls methods to create (a) the cell's diffusion wells, (b) interface-nodes 
 * that connect the cell's substructures (such as the and-plane and input xistors),
 * and (c) creates those substructures.
 * 
 * Each of the methods referenced above adds nodes, arcs and exports to the
 * storage vectors, for use in the object's "text" method.
 * 
 * Note: the only things truly generated in the Cell class are the interface
 * nodes, and diffusion wells.  All other nodes, arcs, and exports are generated
 * in the cell's substructures, and then referenced by the cell.
 */
public class LayCell {

	private String poly;

	private String metal1;

	private String polyM1;

	private String m1M2;

	private Caseinfo info; // provides pertenant parsed case-statement

	// information
	private Paramgetter params; // provides pertenant perameters

	// private Vector layArcs; // stores all arc objects

	// private Vector layExps; // stores all export objects

	// private Vector layNodes; // stores all node objects

	private Vector andPullAndPlane;// stores interface nodes running between

	// the and-pullups and the and-plane

	private Vector inInvAndPlane;// stores interface nodes running between

	// the and plane and the input inverters

	private Vector gndAndPlane; // stores interface nodes running between the

	// ground-line and the and-plane

	private Vector orPlaneAndPlane; // stores interface nodes running between

	// the or-plane and the and-plane

	private Vector andPullPwr; // stores interface nodes running between the

	// and-pullups and the power (Vdd) line

	private Vector inInvPwr; // stores interface nodes running between the

	// input-inverters and the power line

	private Vector outInvPwr; // stores interface nodes running between the

	// output-inverters and power

	private Vector orPullPwr; // stores interface nodes running between the

	// or-pullups and power

	private Vector orPlaneGnd; // stores interface nodes running between the

	// or-plane and ground

	private Vector inInvGnd; // stores interface nodes running between the

	// input-inverters and ground

	private Vector orPullOrPlane; // stores interface nodes running between

	// the or-pullups and the or-plane

	private Vector inInvInExp; // stores the interface nodes running between

	// the input-inverters and the input-exports

	private Vector outInvOutExp; // stores the interface nodes running

	// between the output-inverters and
	// the output-exports

	private Vector pwrPwrExp; // stores the interface node between the

	// power-line and the power-export

	private LayNode gndGndExp; // stores the interface node between the

	// ground-line and the ground-export

	private Vector orOutInv; // stores interface nodes between the or-plane

	// and the output-inverters

	// system parameters obtained via the input Paramgetter object
	private double m1M1Dist;

	private double transistorContactDist;

	private double weakTransistorContactDist;

	private double powerGndWidth;

	private double m1Width;

	private double polyWidth;

	private double weakPolyWidth;

	private double inOutNmosTransistorSize;

	private double wellSurround;

	private int numCases;

	private int numInputs;

	private int numContacts;

	private int numPins;

	private int numPLNodes;

	private int numPmos;

	private int numNmos;

	private int numNWells;

	private int numPWells;

	private int numNets;

	private String expString;

	private String nodeString;

	private String arcString;

	public LayCell(Caseinfo caseInfo, Paramgetter paramGetter) {

		poly = "Polysilicon-1-Pin";
		metal1 = "Metal-1-Pin";
		polyM1 = "Metal-1-Polysilicon-1-Con";
		m1M2 = "Metal-1-Metal-2-Con";

		params = paramGetter;

		expString = "";
		nodeString = "";
		arcString = "";

		numContacts = 0;
		numPins = 0;
		numPLNodes = 0;
		numPmos = 0;
		numNmos = 0;
		numPWells = 0;
		numNWells = 0;
		numNets = 0;

		m1M1Dist = params.getM1ToM1();
		transistorContactDist = params.getTransistorToActiveConnection();
		weakTransistorContactDist = params
				.getWeakTransistorToActiveConnection();
		m1Width = params.getM1Width();
		polyWidth = params.getPolyWidth();
		weakPolyWidth = params.getWeakPolyWidth();
		inOutNmosTransistorSize = params.getNMosInvWidth();
		powerGndWidth = params.getPowerAndGndWide();
		wellSurround = params.getWellSurround();

		info = caseInfo;
		numCases = info.getNumCases();
		numInputs = info.getNumInputBits();

		// layNodes = new Vector(1000,1000);
		// layArcs = new Vector(1000,1000);
		// layExps = new Vector(1000,1000);
		andPullAndPlane = new Vector(100, 100);
		inInvAndPlane = new Vector(100, 100);
		gndAndPlane = new Vector(100, 100);
		orPlaneAndPlane = new Vector(100, 100);
		andPullPwr = new Vector(100, 100);
		inInvPwr = new Vector(100, 100);
		outInvPwr = new Vector(100, 100);
		orPullPwr = new Vector(100, 100);
		orPlaneGnd = new Vector(100, 100);
		inInvGnd = new Vector(100, 100);
		orPullOrPlane = new Vector(100, 100);
		inInvInExp = new Vector(100, 100);
		outInvOutExp = new Vector(100, 100);
		pwrPwrExp = new Vector(1, 1);
		gndGndExp = null;
		orOutInv = new Vector(100, 100);

		makeInterfaces();
		System.out.println("Interfaces");
		makeWells();
		System.out.println("Wells");
		callSubClasses();
		System.out.println("SubClasses");
		makeExports();
		System.out.println("Exports");
	}

	private void makeExports() {
		String textDesc = "D5G2;";
		int inCount = 0;
		int outCount = 0;
		String inName = info.getInBusName();
		int numIns = info.getNumInputBits();
		for (int n = numIns; n > 0; n--) {
			String portID = inName + "[" + String.valueOf(inCount) + "]";
			inCount++;
			String nodeName = ((LayNode) inInvInExp.get(n - 1)).getName();
			String flags = "I";
			LayExp exp = new LayExp(portID, "", textDesc, nodeName, "", flags);
			// layExps.add(exp);
			expString += exp.getExpString();
		}
		String outName = info.getOutBusName();
		int numOuts = info.getNumOutputBits();
		for (int n = numOuts; n > 0; n--) {
			String portID = outName + "[" + String.valueOf(outCount) + "]";
			outCount++;
			String nodeName = ((LayNode) outInvOutExp.get(n - 1)).getName();
			String flags = "O";
			LayExp exp = new LayExp(portID, "", textDesc, nodeName, "", flags);
			// layExps.add(exp);
			expString += exp.getExpString();
		}

		String gndPortId = "gnd";
		String pwrPortId = "vdd";
		String gndNodeName = gndGndExp.getName();
		String pwrNodeName = ((LayNode) pwrPwrExp.get(0)).getName();
		String gndFlag = "G";
		String pwrFlag = "P";

		LayExp gexp = new LayExp(gndPortId, "", textDesc, gndNodeName, "",
				gndFlag);
		expString += gexp.getExpString();
		LayExp pexp = new LayExp(pwrPortId, "", textDesc, pwrNodeName, "",
				pwrFlag);
		expString += pexp.getExpString();
		// layExps.add(gexp);
		// layExps.add(pexp);
	}

	/*
	 * Method: callSubClasses Inputs: none Returns: void Tasks: Calls all
	 * sub-classes in the cell and adds their contribution of arcs and nodes to
	 * the cell's storage vectors for such data.
	 */
	private void callSubClasses() {
		// System.out.println("BeforeAndPin@"+ String.valueOf(numPins));
		AndPlane andPlane = new AndPlane(andPullAndPlane, orPlaneAndPlane,
				inInvAndPlane, gndAndPlane, params, info, numNets, numContacts,
				numPins, numNmos, numPWells);
		// addNodes(andPlane.getNodes());
		// //addArcs(andPlane.getArcs());
		nodeString += andPlane.getNodeString();
		arcString += andPlane.getArcString();

		numNets += andPlane.getNumNets();

		numContacts += andPlane.getNumContacts();

		numPins += andPlane.getNumPins();

		numNmos += andPlane.getNumNmos();

		numPWells += andPlane.getNumPWells();

		// System.out.println("BeforeOrPin@"+ String.valueOf(numPins));
		OrPlane orPlane = new OrPlane(orPlaneAndPlane, orPullOrPlane, orOutInv,
				params, info, numNets, numContacts, numPins, numNmos, numPWells);
		// addNodes(orPlane.getNodes());
		// //addArcs(orPlane.getArcs());
		nodeString += orPlane.getNodeString();
		arcString += orPlane.getArcString();

		numNets += orPlane.getNumNets();

		numContacts += orPlane.getNumContacts();

		numPins += orPlane.getNumPins();

		numNmos += orPlane.getNumNmos();

		numPWells += orPlane.getNumPWells();

		// System.out.println("BeforeAndPullPin@"+ String.valueOf(numPins));
		AndPull andPull = new AndPull(andPullPwr, andPullAndPlane, params,
				numNets, numContacts, numPins, numPmos, numNWells);
		// addNodes(andPull.getNodes());
		// addArcs(andPull.getArcs());
		nodeString += andPull.getNodeString();
		arcString += andPull.getArcString();

		numNets += andPull.getNumNets();

		numContacts += andPull.getNumContacts();

		numPins += andPull.getNumPins();

		numPmos += andPull.getNumPmos();

		numNWells += andPull.getNumNWells();

		// System.out.println("BeforeOrPullPin@"+ String.valueOf(numPins));
		OrPull orPull = new OrPull(orPullPwr, orPullOrPlane, orPlaneGnd,
				params, numNets, numContacts, numPins, numPmos, numNWells);
		// addNodes(orPull.getNodes());
		// addArcs(orPull.getArcs());
		nodeString += orPull.getNodeString();
		arcString += orPull.getArcString();

		numNets += orPull.getNumNets();

		numContacts += orPull.getNumContacts();

		numPins += orPull.getNumPins();

		numPmos += orPull.getNumPmos();

		numNWells += orPull.getNumNWells();
		// System.out.println("BeforeInInvPin@"+ String.valueOf(numPins));
		InInv inInv = new InInv(inInvPwr, inInvGnd, inInvInExp, inInvAndPlane,
				params, numNets, numContacts, numPins, numPmos, numNmos);
		// addArcs(inInv.getArcs());
		// addNodes(inInv.getNodes());
		nodeString += inInv.getNodeString();
		arcString += inInv.getArcString();

		numNets += inInv.getNumNets();

		numContacts += inInv.getNumContacts();

		numPins += inInv.getNumPins();

		numPmos += inInv.getNumPmos();

		numNmos += inInv.getNumNmos();

		// System.out.println("BeforeOutInvPin@"+ String.valueOf(numPins));
		OutInv outInv = new OutInv(outInvPwr, orOutInv, outInvOutExp, params,
				numNets, numContacts, numPins, numPmos, numNmos);
		// addNodes(outInv.getNodes());
		// addArcs(outInv.getArcs());
		nodeString += outInv.getNodeString();
		arcString += outInv.getArcString();

		numNets += outInv.getNumNets();

		numContacts += outInv.getNumContacts();

		numPins += outInv.getNumPins();

		numPmos += outInv.getNumPmos();

		numNmos += outInv.getNumNmos();

		// System.out.println("BeforeGndPin@"+ String.valueOf(numPins));
		Gnd gnd = new Gnd(inInvGnd, gndAndPlane, orPlaneGnd, gndGndExp, params,
				numNets, numPins, numPWells, numContacts);
		// addNodes(gnd.getNodes());
		// addArcs(gnd.getArcs());
		nodeString += gnd.getNodeString();
		arcString += gnd.getArcString();

		numNets += gnd.getNumNets();

		numPins += gnd.getNumPins();

		numPWells += gnd.getNumPWells();

		numContacts += gnd.getNumContacts();

		// System.out.println("BeforePwrPin@"+ String.valueOf(numPins));
		Pwr pwr = new Pwr(orPullPwr, andPullPwr, inInvPwr, outInvPwr,
				pwrPwrExp, params, numNets, numPins, numNWells, outInvOutExp);
		// addNodes(pwr.getNodes());
		// addArcs(pwr.getArcs());
		nodeString += pwr.getNodeString();
		arcString += pwr.getArcString();

		numNets += pwr.getNumNets();

		numPins += pwr.getNumPins();

		numNWells += pwr.getNumNWells();

	}

	/*
	 * private void addArcs(Vector arcs) { int numAdds = arcs.size(); for (int n =
	 * 0; n < numAdds; n++) { layArcs.add(arcs.get(n)); } } private void
	 * addNodes(Vector nodes) { int numAdds = nodes.size(); for (int n = 0; n <
	 * numAdds; n++) { layNodes.add(nodes.get(n)); } }
	 */

	/*
	 * Method: makeInterfaces Inputs: none Returns: void Tasks: uses the cell's
	 * parameters (params) and case-statement data (info) to generate cell nodes
	 * that run between cell sub-systems, including: 1. and-pullups/and-plane 2.
	 * in-inverters/and-plane 3. and-plane/gnd 4. and-plane/or-plane 5.
	 * power(vdd)/and-pullups 6. power/in-inverters 7. power/out-inverters 8.
	 * power/or-pullups 9. gnd/or-plane 10. gnd/in-inverters 11.
	 * or-plane/or-pullups 12. in-inverters/input-exports 13.
	 * out-inverters/output-exports 14. power/pwr-export 15. gnd/gnd-export 16.
	 * or-plane/out-inverters This method finds the x and y coordinates of the
	 * first node (lower left) in the interface, and then calls the
	 * getInterfaceVec method to create the vector containing all of the
	 * interface nodes.
	 */
	private void makeInterfaces() {

		// no interval: for the axis of alignment of the given nodes
		int inline = 0;

		// interval between parallel M1 columns
		double interval1 = params.getM1Width() + params.getM1ToM1();

		// interval between polysilicon columns in the and-plane and inverters
		double interval2 = params.getPolyWidth()
				+ params.getTransistorToActiveConnection()
				+ params.getM1Width() + params.getPolyToDiff();

		/*
		 * Format of method is: interfaceVector( (String) type of metal of
		 * interface node, (double) first Y-coordinate, (double) first X
		 * coordinate, (int) interval of seperation along the Y-axis, (int)
		 * interval of seperation along X-axis, (int) number of interface nodes)
		 * method "aInterfaceVector(...) follows similar formatting, but is used
		 * for situations where the interval between nodes is inconsistant.
		 * There is an additional (int)-arguement used to specify the alternate
		 * spacing, and an (int)-arguement specifying the ratio of initial to
		 * alternate spacing.
		 */

		// //System.out.print("\nand-pulldowns / and-plane");
		andPullAndPlane = interfaceVector(metal1, andPullAndPlaneLow(),
				andPullAndPlaneLeft(), interval1, inline,
				numAndPullAndPlaneInterfaceNodes());

		// ////System.out.print("\ninput-inverters / and-plane");
		inInvAndPlane = interfaceVector(poly, inInvAndPlaneLow(),
				inInvAndPlaneLeft(), inline, interval2,
				numInInvAndPlaneInterfaceNodes());

		// //System.out.print("\nground / and-plane");
		gndAndPlane = interfaceVector(m1M2, gndAndPlaneLow(),
				gndAndPlaneLeft(), 3 * interval1, inline,
				numGndAndPlaneInterfaceNodes());

		// //System.out.print("\nor-plane / and-plane");
		orPlaneAndPlane = aInterfaceVector(polyM1, orPlaneAndPlaneLow(),
				orPlaneAndPlaneLeft(), 2 * interval1, inline, interval1, 1,
				numOrPlaneAndPlaneInterfaceNodes());

		// //System.out.print("\nand-pulldowns / power");
		andPullPwr = aInterfaceVector(metal1, andPullPwrLow(),
				andPullPwrLeft(), 2 * interval1, inline, interval1, 1,
				numAndPullPwrInterfaceNodes());

		// //System.out.print("\ninput-inverters / power");
		inInvPwr = interfaceVector(metal1, inInvPwrLow(), inInvPwrLeft(),
				inline, 2 * interval2, numinInvPwrInterfaceNodes());

		// //System.out.print("\noutput-inverters / power");
		outInvPwr = interfaceVector(metal1, outInvPwrLow(), outInvPwrLeft(),
				inline, 2 * interval2 + interval1, numOutInvPwrInterfaceNodes());

		// //System.out.print("\nor-pullups / power");
		orPullPwr = aInterfaceVector(metal1, orPullPwrLow(), orPullPwrLeft(),
				inline, 2 * interval2, interval1, 1,
				numOrPullPwrInterfaceNodes());

		// //System.out.print("\nor-plane / ground");
		orPlaneGnd = interfaceVector(m1M2, orPlaneGndLow(), orPlaneGndLeft(),
				inline, 2 * interval2 + interval1,
				numOrPlaneGndInterfaceNodes());

		// //System.out.print("\ninput-inverters / ground");
		inInvGnd = interfaceVector(metal1, inInvGndLow(), inInvGndLeft(),
				inline, 2 * interval2, numInInvGndInterfaceNodes());

		// //System.out.print("\nor-pulldowns / or-plane");
		orPullOrPlane = aInterfaceVector(metal1, orPullOrPlaneLow(),
				orPullOrPlaneLeft(), inline, interval2, interval1, 2,
				numOrPullOrPlaneInterfaceNodes());

		// //System.out.print("\ninput-inverters / input-exports");
		inInvInExp = interfaceVector(m1M2, inInvInExpLow(), inInvInExpLeft(),
				inline, 2 * interval2, numInInvInExpInterfaceNodes());

		// //System.out.print("\noutput-inverters / output-exports");
		outInvOutExp = aInterfaceVector(m1M2, outInvOutExpLow(),
				outInvOutExpLeft(), inline, 2 * interval2, interval1, 1,
				numOutInvOutExpInterfaceNodes());

		// //System.out.print("\npower - power-export");
		pwrPwrExp = interfaceVector(m1M2, pwrPwrExpLow(), pwrPwrExpLeft(),
				inline, inline, 1);

		// //System.out.print("\nground / ground-export");
		gndGndExp = makeGndExpNode();

		// //System.out.print("\nor-plane / output-inverters");
		orOutInv = aInterfaceVector(metal1, orOutInvLow(), orOutInvLeft(),
				inline, interval2, interval1, 2, numOrOutInvInterfaceNodes());
	}

	private LayNode makeGndExpNode() {
		String type = "Metal-1-Metal-2-Con";
		double mW = params.getPowerAndGndWide();
		String name = "contact@" + String.valueOf(numContacts);
		numContacts++;
		// System.out.println("pin@"+ String.valueOf(numPins));
		double xVal = gndGndExpLeft();
		double yVal = gndGndExpLow();
		String flags = "V";
		LayNode node = new LayNode(type, name, "", xVal, yVal, mW, mW, "",
				flags);
		nodeString += node.getText();
		// layNodes.add(node);
		return node;
	}

	/*
	 * Method: numOrOutInvInterfaceNodes() Inputs: none Returns: (int) The
	 * number of nodes linking the or-plane and the output inverters Tasks: none
	 */

	private int numOrOutInvInterfaceNodes() {
		return (int) (Math.ceil(1.5 * info.getNumOutputBits()));
	}

	/*
	 * Method: orOutInvLeft() Inputs: none Returns: (double) the X position of
	 * the leftmost node linking the or-plane and the output inverters Tasks:
	 * none
	 */
	private double orOutInvLeft() {
		return gndGndExpLeft() + (.5 * powerGndWidth) + m1M1Dist
				+ (.5 * m1Width);
	}

	/*
	 * Method: orOutInvLow() Inputs: none Returns: (double) value of the
	 * Y-position of the first (low-left) node linking the or-plane and the
	 * output inverters Tasks: none
	 */
	private double orOutInvLow() {
		return (.5 * powerGndWidth) + m1M1Dist + (2 * inOutNmosTransistorSize)
				+ m1M1Dist + m1Width + m1M1Dist + m1Width + m1M1Dist
				+ inOutNmosTransistorSize + m1M1Dist + (.5 * powerGndWidth);
	}

	/*
	 * Method: gndGndExpLeft() Inputs: None Returns: (double) X-axis location of
	 * the ground export Tasks: None
	 */
	private double gndGndExpLeft() {
		return wellSurround + (1.5 * m1Width) + weakTransistorContactDist
				+ weakPolyWidth + weakTransistorContactDist + m1Width
				+ (2 * wellSurround) + numInputs
				* (m1Width * 2 + transistorContactDist * 4 + 2 * polyWidth)
				+ m1Width + m1M1Dist + (.5 * m1Width);
	}

	/*
	 * Method: gndGndExpLow() Inputs: none Returns: (double) Y-axis location of
	 * the ground export Tasks: none
	 */
	private double gndGndExpLow() {
		return orPullPwrLow() + powerGndWidth + m1M1Dist;
	}

	/*
	 * Method: pwrPwrExpLow() Inputs: none Returns: (double) value of the
	 * Y-position for the power export. It is placed at (0,0), so the value will
	 * be 0.0 Tasks: none
	 */
	private double pwrPwrExpLow() {
		return 0.0;
	}

	/*
	 * Method: pwrPwrExpLeft() Inputs: none Returns: (double) value of the
	 * X-position for the power export. It is placed at (0,0), so the value will
	 * be 0.0 Tasks: none
	 */
	private double pwrPwrExpLeft() {
		return 0.0;
	}

	/*
	 * Method: outInvOutExpLow() Inputs: none Returns: (double) value of the
	 * Y-position of the first output node Tasks:
	 */
	private double outInvOutExpLow() {
		return (.5 * powerGndWidth) + m1M1Dist + (2 * inOutNmosTransistorSize)
				+ m1M1Dist + (.5 * m1Width);
	}

	/*
	 * Method: outInvOutExpLeft() Inputs: None Returns: (double) X-position of
	 * the lower-left node joining the output-Inverters and the output-exports.
	 * Tasks: Calculation and return of the value, from circuit parameters
	 */
	private double outInvOutExpLeft() {
		return orOutInvLeft();
	}

	/*
	 * Method: numOutInvOutExpInterfaceNodes() Inputs: None Returns: (int) The
	 * number of nodes linnking the output-inverters and the Output Exports.
	 * (This is simply the number of output bits/exports) Tasks:
	 */
	private int numOutInvOutExpInterfaceNodes() {
		return info.getNumOutputBits();
	}

	/*
	 * Method: inInvInExpLeft() Inputs: none Returns: (double) value of the
	 * X-position of the lower left node joining the input-inverters and the
	 * input-exports. Tasks: Calculation and return of the value from cell
	 * parameters
	 */
	private double inInvInExpLeft() {
		return 4 + (1.5 * m1Width) + weakTransistorContactDist + weakPolyWidth
				+ weakTransistorContactDist + m1Width + (2 * wellSurround)
				+ (.5 * m1Width);
	}

	/*
	 * Method: inInvInExpLow() Inputs: none Returns: (double) the Y-value of the
	 * lower left input export Tasks: calculate and return the value
	 */
	private double inInvInExpLow() {
		return (.5 * powerGndWidth) + m1M1Dist + (2 * inOutNmosTransistorSize)
				+ m1M1Dist + (.5 * m1Width);
	}

	/*
	 * Method: numInInvInExpInterfaceNodes() Inputs: None Returns: (int) the
	 * number of nodes linking the Input-inverters and the input exports (this
	 * is just the number of input bits) Tasks: calculation and return of the
	 * value, using available cell information.
	 */
	private int numInInvInExpInterfaceNodes() {
		return info.getNumInputBits();
	}

	/*
	 * Method: orPullOrPlaneLeft() Inputs: none Returns: (double) the X-position
	 * of the lower left node connecting the or-pullups and the or-plane Tasks:
	 * Calculate and return the value, using cell parameters
	 */
	private double orPullOrPlaneLeft() {
		return orOutInvLeft();
	}

	/*
	 * Method: orPullOrPlaneLow Inputs: None Returns: (double) Y-coordinate of
	 * the node linking the or-pullups and the or-plane. Tasks: Calculate and
	 * return the value, using cell parameters
	 */
	private double orPullOrPlaneLow() {
		return (.5 * powerGndWidth) + m1M1Dist + (2 * inOutNmosTransistorSize)
				+ m1M1Dist + m1Width + m1M1Dist + m1Width + m1M1Dist
				+ inOutNmosTransistorSize + m1M1Dist + powerGndWidth + m1M1Dist
				+ (m1Width * Math.ceil(numCases * 1.5))
				+ (m1M1Dist * (Math.ceil(numCases * 1.5) - 1)) + wellSurround
				+ 4;
	}

	/*
	 * Method: numOrPullOrPlaneInterfaceNodes() Inputs: None Returns: (int) the
	 * number of nodes linking the or-pullups and the or-plane. Tasks: calculate
	 * and return the given value, using available cell information
	 */
	private int numOrPullOrPlaneInterfaceNodes() {
		return (int) (Math.ceil(1.5 * info.getNumOutputBits()));
	}

	/*
	 * Method: inInvGndLow() Inputs: none Returns: (double) Y-coordinate of the
	 * lower-left node connecting the input-inverters and Ground-line Tasks:
	 * calculation and return of the value, using available cell parameters
	 */
	private double inInvGndLow() {
		return orOutInvLow();
	}

	/*
	 * Method: inInvGndLeft() Inputs: none Returns: (double) X-coordinate of the
	 * lower-left node connecting the input Inverters and the Ground-line.
	 * Tasks: Calculation and return of the value, based upon available cell
	 * parameters
	 */
	private double inInvGndLeft() {
		return inInvInExpLeft();
	}

	/*
	 * Method: numInInvGndInterfaceNodes Inputs: None Returns: (int) the number
	 * of nodes linking the input-inverters and the ground- line Tasks:
	 * Calculation/return of the value based on available cell info.
	 */
	private int numInInvGndInterfaceNodes() {
		return info.getNumInputBits();
	}

	/*
	 * Method: orPlaneGndLeft Inputs: none Returns: (double) X-coordinate of the
	 * low-left node connecting the or-plane and the ground-line. Tasks:
	 * Calculation/return of the value, based on available cell parameters.
	 */
	private double orPlaneGndLeft() {
		return orPullOrPlaneLeft() + (.5 * m1Width)
				+ (2 * transistorContactDist) + polyWidth + (.5 * m1Width);
	}

	/*
	 * Method: orPlaneGndLow() Inputs: none Returns: (double) Y-coordinate of
	 * the low-left node connecting the or-plane and the ground-line. Tasks:
	 * Calculation/return of the value, based upon cell parameters
	 */
	private double orPlaneGndLow() {
		return gndGndExpLow();
	}

	/*
	 * Method: numOrPlaneGndInterfaceNodes() Inputs: None Returns: (int) the
	 * number of nodes linking the ground-line to the or-plane Tasks: None
	 */
	private int numOrPlaneGndInterfaceNodes() {
		return (int) (Math.ceil(.5 * info.getNumOutputBits()));
	}

	/*
	 * Method: numOrPullPwrInterfaceNodes() Inputs: None Returns: (int) the
	 * number of nodes linking the or-pullups to the power line. Tasks: None
	 */
	private int numOrPullPwrInterfaceNodes() {
		return info.getNumOutputBits();
	}

	/*
	 * Method: orPullPwrLeft() Inputs: none Returns: (double) X-coordinate of
	 * the low-left node connecting the or-pullups and the power-line Tasks:
	 * Calculation of the value, using available cell parameters.
	 */
	private double orPullPwrLeft() {
		return orPullOrPlaneLeft();
	}

	/*
	 * Method: orPullPwrLow() Inputs: None Returns: (double) Y-coordinate of the
	 * low-left node connecting the or-pullups and the power-line Tasks:
	 * Calculation/return of the value, based on available cell parameters.
	 */
	private double orPullPwrLow() {
		return orPullOrPlaneLow() + (2 * wellSurround) + m1Width
				+ weakTransistorContactDist + weakPolyWidth
				+ weakTransistorContactDist + m1Width;
	}

	/*
	 * Method: numOutInvPwrInterfaceNodes() Inputs: none Returns: (int) the
	 * number of nodes joining the output inverter with the power line Tasks:
	 * none
	 */
	private int numOutInvPwrInterfaceNodes() {
		return (int) (Math.ceil(info.getNumOutputBits() * .5));
	}

	/*
	 * Method: outInvPwrLeft() Inputs: none Returns: (double) X-coordinate of
	 * the low-left node connecting the output-inverters to the power-line
	 * Tasks: calculation/return of the value, based on available cell
	 * parameters
	 */
	private double outInvPwrLeft() {
		return outInvOutExpLeft() + (.5 * m1Width) + transistorContactDist
				+ polyWidth + transistorContactDist + (.5 * m1Width);
	}

	/*
	 * Method: outInvPwrLow() Inputs: none Returns: (double) Y-coordinate of the
	 * low-left node connecting the output-inverters and the power-line Tasks:
	 * calculation/return of the value, using available cell parameters. (Should
	 * be zero)
	 */
	private double outInvPwrLow() {
		return 0.0;
	}

	/*
	 * Method: numinInvPwrInterfaceNodes() Inputs: none Returns: (int) number of
	 * nodes joining the input-inverters to Power Tasks: none
	 */
	private int numinInvPwrInterfaceNodes() {
		return info.getNumInputBits();
	}

	/*
	 * Method: inInvPwrLeft() Inputs: none Returns: (double) X-coordinate of the
	 * low-left node linking the input-inverters and the power-line Tasks:
	 * calculate/return the value, based upon available cell parameters.
	 */
	private double inInvPwrLeft() {
		return inInvInExpLeft();
	}

	/*
	 * Method: inInvPwrLow() Inputs: none Returns: (double) Y-value of the
	 * low-left node connecting the input-inverters and the power-line (should
	 * be 0) Tasks: Calculate/return the value.
	 */
	private double inInvPwrLow() {
		return 0.0;
	}

	/*
	 * Method: numAndPullPwrInterfaceNodes() Inputs: none Returns: (int) the
	 * number of nodes joining the and-pulldowns to the power-line Tasks: none
	 */
	private int numAndPullPwrInterfaceNodes() {
		return info.getNumCases();
	}

	/*
	 * Method: andPullPwrLeft() Inputs: none Returns: (double) X-coordinate of
	 * the low-left node connecting the and-pullups and the power-line. (Should
	 * be 0) Tasks: calculate/return the given value, based on available cell
	 * parameters
	 */
	private double andPullPwrLeft() {
		return 0.0;
	}

	/*
	 * Method: andPullPwrLow Inputs: none Returns: (double) Y-coordinate of the
	 * low-left node connecting the and-pullups and the powerline Tasks:
	 * calculate/return the given value, based on available cell parameters
	 */
	private double andPullPwrLow() {
		return inInvGndLow() + (.5 * powerGndWidth) + m1M1Dist + (.5 * m1Width);
	}

	/*
	 * Method: aInterfaceVector Inputs: (String) nodeType: the type of node
	 * (via, pin, etc...) (double) firstY: the y-coordinate of the first node to
	 * be created (double) firstX: the x-coordinate of the first node to be
	 * created (int) yInterval: the interval by which the node will "jump" along
	 * the y-axis (int) xInterval: the interval by which the node will "jump"
	 * along the x-axis (int) alternateInterval: an alternate interval (int) r
	 * the ratio of normal (x or y) to alternate intervals. For example, if r is
	 * 2, the node will jump 2x at the normal interval, and then once at the
	 * alternate interval. (int) numNodes: the number of nodes to be generated.
	 * Returns: Vector object containing all nodes generated in the method
	 * Tasks: generates/returns a Vector of "numNodes" nodes, spaced at
	 * intervals specified by (xInterval or yInterval) and altInterval, at the
	 * ratio specified by r. Adds each node created to the Cell's LayNodes
	 * Vector.
	 */

	/*
	 * Method: orPlaneAndPlaneLeft Inputs: none Returns: (double) X-coordinate
	 * of the low-left node connecting the or-plane and the and-plane Tasks:
	 * calculation/return of the value, based on available cell parameters
	 */
	private double orPlaneAndPlaneLeft() {
		return gndGndExpLeft();
	}

	/*
	 * Method: orPlaneAndPlaneLow Inputs: none Returns: (double) Y-coordinate of
	 * the low-left node connecting the or-plane and the and-plane Tasks:
	 * calculate/return the given value, based on available cell parameters
	 */
	private double orPlaneAndPlaneLow() {
		return andPullPwrLow();
	}

	/*
	 * Method: numOrPlaneAndPlaneInterfaceNodes() Inputs: none Returns: (int)
	 * the number of nodes joining the or-plane and the and-plane Tasks: None
	 */
	private int numOrPlaneAndPlaneInterfaceNodes() {
		return info.getNumCases();
	}

	/*
	 * Method: numGndAndPlaneInterfaceNodes() Inputs: none Returns: (int) the
	 * number of nodes joining the ground line and the and-plane Tasks:
	 * calculate/return the given value, using available cell information
	 */
	private int numGndAndPlaneInterfaceNodes() {
		// System.out.println(String.valueOf((int) (Math.ceil(.5 *
		// info.getNumCases()))));
		return (int) (Math.ceil(.5 * info.getNumCases()));
	}

	/*
	 * Method: gndAndPlaneLow Inputs: nonoe Returns: (double) Y-coordinate of
	 * the low-left node connecting the and-plane to the ground-line Tasks:
	 * calculate/return the given value, based on available cell information
	 */
	private double gndAndPlaneLow() {
		return andPullPwrLow() + (.5 * m1Width) + m1M1Dist + (.5 * m1Width);
	}

	/*
	 * Method: gndAndPlaneLeft Inputs: none Returns: (double) X-coordinate of
	 * the lower-left node connecting the ground-line to the and-plane. Tasks:
	 * calculation/return of the given value, based on available cell parameters
	 */
	private double gndAndPlaneLeft() {
		return gndGndExpLeft();
	}

	/*
	 * Method: numInvAndPlaneInterfaceNodes Inputs: none Returns: (int) the
	 * number of nodes linking the and-plane to the input-inverters Tasks:
	 * calculate/return the given value, based on available cell information.
	 */
	private int numInInvAndPlaneInterfaceNodes() {
		return 2 * info.getNumInputBits();
	}

	/*
	 * Method: inInvAndPlaneLeft Inputs: none Returns: (double) X-coordinate of
	 * the lower-left node linking the input-inverters and the and-plane Tasks:
	 * calculation/return of the given value, based on available cell parameters
	 */
	private double inInvAndPlaneLeft() {
		return inInvInExpLeft() + (.5 * m1Width) + transistorContactDist
				+ (.5 * polyWidth);
	}

	/*
	 * Method: inInvAndPlaneLow Inputs: none Returns: (double) Y-coordinate of
	 * the lower-left node linking the input-inverters and the and-plane Tasks:
	 * calculation/return of the given value, based on available cell parameters
	 */
	private double inInvAndPlaneLow() {
		return inInvGndLow();
	}

	/*
	 * Method: andPullAndPlaneLeft Inputs: none Returns: (double) X-coordinate
	 * of the lower-left node linking the and-pullups and the and-plane Tasks:
	 * calculation/return of the given value, based on available cell parameters
	 */
	private double andPullAndPlaneLeft() {
		return inInvInExpLeft() - ((.5 * m1Width) + m1M1Dist + (.5 * m1Width));
	}

	/*
	 * Method: andPullAndPlaneLow Inputs: none Returns: (double) Y-coordinate of
	 * the lower-left node linking the and-pullups and the and-plane Tasks:
	 * calculation/return of the given value, based on available cell parameters
	 */
	private double andPullAndPlaneLow() {
		return andPullPwrLow();
	}

	/*
	 * Method: numAndPullAndPlaneInterfaceNodes() Inputs: none Returns: (int)
	 * the number of nodes linking the and-pullups with the and-plane Tasks:
	 * Calculate/return the given value, based on available cell information
	 */
	private int numAndPullAndPlaneInterfaceNodes() {
		return (int) (Math.ceil(info.getNumCases() * 1.5));
	}

	/*
	 * Method: interfaceVector Inputs: (String) nodeType: The type of node being
	 * generated (double) firstY: The Y-coordinate of the first node to be
	 * generated (double) firstX: The X-coordinate of the first node to be
	 * generated (int) yInterval: The y-incrementation of subsequent nodes (int)
	 * xInterval: The x-incrementation of subsequent nodes (int) numNodes: The
	 * number of nodes to be made Returns: Vector object containing all
	 * interface nodes created by the method Tasks: Generates a vector of
	 * layNode objects based on the input arguements, adding them to the Cell's
	 * layNodes Vector and to the returned vector as they are created.
	 */
	private Vector interfaceVector(String nodeType, double firstY,
			double firstX, double interval1, double xInterval, int numNodes) {
		String name = null;
		String type = nodeType;
		double width = 0;
		LayNode node = null;

		double y = firstY;
		double x = firstX;
		double yJump = interval1;
		double xJump = xInterval;
		int numberOfNodes = numNodes;

		Vector nodes = new Vector(100, 100);

		for (int n = 0; n < numberOfNodes; n++) {
			if (type.equals(polyM1) || type.equals(m1M2)) {
				name = "contact@" + String.valueOf(numContacts);
				node = new LayNode(type, name, "", x, y, m1Width, m1Width, "",
						"V");
				numContacts++;
			} else {
				name = "pin@" + String.valueOf(numPins);
				if (type.equals(poly)) {
					width = polyWidth;
				} else
					width = m1Width;
				node = new LayNode(type, name, "", x, y, width, width, "", "V");
				numPins++;
			}
			// ////System.out.print(node.getText());
			nodeString += node.getText();
			nodes.add(node);
			// layNodes.add(node);
			x += xJump;
			y += yJump;
		}
		return nodes;
	}

	private Vector aInterfaceVector(String nodeType, double firstY,
			double firstX, double d, double interval2, double interval1, int r,
			int numNodes) {

		String name = null;
		String type = nodeType;
		double y = firstY;
		double x = firstX;
		double yJump = d;
		double xJump = interval2;
		double altJump = interval1;
		int ratio = r;
		int numberOfNodes = numNodes;

		int count = 0; // used for keeping track of jumping to ensure proper
		// intervals

		double width = 0;// used in node creation: the width/height of the
		// node

		// node creation
		Vector nodes = new Vector(1, 1);
		LayNode node = null;

		for (int n = 0; n < numberOfNodes; n++) {
			if (type.equals(polyM1) || type.equals(m1M2)) {
				name = "contact@" + String.valueOf(numContacts);
				node = new LayNode(type, name, "", x, y, m1Width, m1Width, "",
						"V");
				numContacts++;
			} else {
				name = "pin@" + String.valueOf(numPins);
				if (type.equals(poly)) {
					width = polyWidth;
				} else
					width = m1Width;
				node = new LayNode(type, name, "", x, y, width, width, "", "V");
				numPins++;
			}

			// //System.out.print(node.getText());
			nodeString += node.getText();
			nodes.add(node);
			// layNodes.add(node);

			// creating proper jump intervals
			if (yJump == 0) {
				if (count < ratio) {
					x += xJump;
					count++;
				} else {
					x += altJump;
					count = 0;
				}
			} else {
				if (count < ratio) {
					y += yJump;
					count++;
				} else {
					y += altJump;
					count = 0;
				}
			}
		}
		return nodes;
	}

	/*
	 * Method: makeWells Inputs: none Returns: void Tasks: creates the various
	 * wells for the cell
	 */
	private void makeWells() {
		String pWell = "P-Well-Node";
		String nWell = "N-Well-Node";
		createWell(nWell, leftLeft(), rightLeft(), topLeft(), bottomLeft());
		createWell(nWell, leftBottom(), rightBottom(), topBottom(),
				bottomBottom());
		createWell(nWell, leftTop(), rightTop(), topTop(), bottomTop());
		createWell(pWell, rightLeft(), rightTop(), bottomTop(), topBottom());
	}

	private double topLeft() {
		return ((LayNode) andPullPwr.get(andPullPwr.size() - 1)).getY()
				+ (.5 * m1Width) + wellSurround;
	}

	private double leftLeft() {
		return ((LayNode) andPullPwr.get(0)).getX()
				- ((.5 * m1Width) + wellSurround);
	}

	private double bottomLeft() {
		return ((LayNode) andPullPwr.get(0)).getY()
				- ((.5 * m1Width) + wellSurround);
	}

	private void createWell(String well, double left, double right, double top,
			double bottom) {
		String type = well;
		double centerX = .5 * (right + left);
		double centerY = .5 * (top + bottom);
		double width = right - left;
		double height = top - bottom;

		String name = "plnode@" + String.valueOf(numPLNodes);
		numPLNodes++;

		String flag = "AV";

		LayNode node = new LayNode(type, name, "", centerX, centerY, width,
				height, "", flag);
		nodeString += node.getText();
		// layNodes.add(node);
	}

	private double leftBottom() {
		return inInvInExpLeft() - ((.5 * m1Width) + wellSurround);
	}

	private double rightBottom() {
		return rightTop();
	}

	private double bottomBottom() {
		return -((.5 * m1Width) + wellSurround);
	}

	private double leftTop() {
		return orPullPwrLeft() - ((.5 * m1Width) + wellSurround);
	}

	private double rightLeft() {
		return inInvGndLeft() - ((.5 * m1Width) + wellSurround);
	}

	private double bottomTop() {
		return orPullOrPlaneLow();
	}

	private double topTop() {
		return orPullPwrLow() + ((.5 * m1Width) + wellSurround);
	}

	private double rightTop() {
		return ((LayNode) orPullOrPlane.get(orPullOrPlane.size() - 1)).getX()
				+ ((.5 * m1Width) + wellSurround);
	}

	private double topBottom() {
		return inInvInExpLow() + (.5 * m1Width);
	}

	/*
	 * Method: text Inputs: None Returns: String representing the cell's nodes,
	 * arcs, and exports, formatted to be read by electric. Tasks: Uses the
	 * cell's storage-vectors (containing node, arc and export objects) to
	 * generate and return a text file that will be read- able by electric.
	 */
	public String text() {

		System.out.println("textStart");
		Date date = new Date();// obtains the a Date-object of the present
		// date and time, using epoch time formatting
		Long timestamp = new Long(date.getTime());// Converts the Date-object
		// into a Long (NOT PRIMATIVE) object.
		String ts = timestamp.toString();// Converts the Long object to a
		// string

		// Generates the layout "header"
		String header = "\n\n# Cell " + params.getCellName() + "_Cell;1{lay}"
				+ "\nC" + params.getCellName() + "_Cell;1{lay}||" + "mocmos" + "|"
				+ ts + "|" + ts + "|E\n" + "N" + "generic:Facet-Center" + "|"
				+ "art@0" + "|" + "|" + "0" + "|" + "0" + "|" + "0" + "|" + "0"
				+ "|" + "|" + "AV\n";

		return header + nodeString + arcString + expString + "\nX";

		/*
		 * String cellString = null;// Will contain the cell's text data //
		 * obtains the number of nodes, arcs, and exports int numLayNodes =
		 * layNodes.size(); int numLayArcs = layArcs.size(); int numLayExps =
		 * layExps.size(); // adds node lines to the cellString file
		 * System.out.println("Check1"); for (int n = 0; n < numLayNodes; n++) {
		 * cellString += ((LayNode) (layNodes.get(n))).getText(); }
		 * System.out.println("\nmid1"); // adds arc lines to the cellString
		 * file System.out.println("Check2 " +numLayArcs); for (int n = 0; n <
		 * numLayArcs; n++) { // System.out.println("Check2" +n+" / "
		 * +numLayArcs); cellString += ((LayArc) (layArcs.get(n))).getText(); }
		 * System.out.println("\nmid2"); // adds export lines to the cellString
		 * file System.out.println("Check1"); for (int n = 0; n < numLayExps;
		 * n++) { // System.out.println("Check2" +n+" / " +numLayExps);
		 * cellString += ((LayExp) (layExps.get(n))).getText(); } // ends the
		 * layout body cellString = cellString + "\nX";
		 * System.out.println("textEnd"); return cellString;
		 */
	}

}
